# -*- coding: UTF-8 -*-
"""
Aggrégateur de fichiers créole
"""

from os.path import isfile, isdir, join
from os import listdir
import ConfigParser
import sys
import traceback

from creole.log import logger
from creole.error import *
from creole.dom_parser import *
from creole.template import Template
from creole.typeole import *
from creole import eosfunc
from creole import dico
from creole import config

class EoleDict:
    """Aggrégateur de fichier eol
    """

    def __init__(self, templatedir='/etc/eole/template', logfile=None):
        """
        files: nom long des fichiers cibles
        variables: dictionnaire des variables
        Les variables sont empilées dans des listes.

        @param templatedir: répertoire des fichiers template
        @param log: instance de logger
        """
        self.store = ConfigParser.ConfigParser()
        self.files = {}
        self.filelists = {}
        self.variables = {}
        self.families = {} # all families
        self.normal_families = [] # only non expert mode families
        self.family_order = []
        self.constraints = {}
        self.orphan_constraints = {'checks':{}, 'fills':{}, 'auto':{}}
        self.auto_vars = {}
        self.groups = {}
        self.helps = {'variables':{}, 'families':{}}
        self.separators = {}
        self.templatedir = templatedir
        self.log = logger
        if logfile == None:
            self.log.set_logfile()
        else:
            self.log.set_logfile(logfile)

    def read(self, filename):
        """Aggregation d'un fichier dictionnaire
        @param filename: fichier de configuration 
        """
        try:
            # on vérifie que le fichier existe
            if not isfile(filename):
                raise FileNotFound, "le fichier %s n'existe pas" %  filename

            #parsing du fichier
            new_files, new_variables, new_families, new_order, new_constraints, new_groups, helps, separators = parse_xml_file(filename)
            self.log.info('lecture du fichier  %s ' % filename)
            
        except Exception, err:
            self.log.error(traceback.format_exc())
            raise ConfigError, "erreur au chargement d'une configuration : %s" % str(err)

        return self._read(new_files, new_variables, new_families, new_order, 
                          new_constraints, new_groups, helps, separators)

    def import_values(self, filename):
        """import de valeurs depuis un fichier creole1
        """
        errors = []
        warnings = []
        try:
            if not isfile(filename):
                raise FileNotFound, "le fichier %s n'existe pas" %  filename
        except Exception, err:
            raise ConfigError, "erreur au chargement d'une configuration : %s" % err
        # chargement du dictionnaire
        old_dict = dico.DicoEole(filename)
        # essai de détection du module Eole d'origine
        module = dico.detect_module(old_dict.liste_vars.keys())
        if module ==  '':
            raise ConfigError, "erreur de détection du type de module"
        # chargement du dictionnaire des conversion des variables 1 -> 2
        dict_convert = dico.read_convert(join(config.datadir, 'convert', '%s.csv' % module))
        # import des valeurs
        for newvar in dict_convert.keys():
            val = []
            # la valeur doit être récupérée dans l'ancienne conf
            for old_var in dict_convert[newvar]:
                try:
                    old_val = old_dict.get_value(old_var)
                except Exception, e:
                    # variable non renseignée dans l'ancien dictionnaire
                    old_val = ''
                if old_val != '':
                    val.append(old_val)
                # insertion de la valeur
                try:
                    self.set_value(newvar, val)
                except KeyError, e:
                    errors.append('variable inconnue :'+str(e))
                except Exception, e:
                    errors.append(str(e))

        # on recalcule les valeurs pour toutes les autres variables
        for family in self.family_order:
            for newvar in self.families[family]['vars']:
                if newvar not in dict_convert.keys():
                    try:
                        val = self.get_value(newvar)
                        self.set_value(newvar, val)
                    except TypeEoleError, e:
                        warnings.append(str(e))
            return errors, warnings

    def read_string(self, xml_string):
        """Aggregation d'un fichier dictionnaire
        @param xml_string: fichier de configuration 
        """
        try:
            if not xml_string.startswith("<?xml"):
                raise FileNotFound, "le dictionnaire doit être au format xml"
            #parsing de la chaîne
            new_files, new_variables, new_families, new_order, new_constraints, new_groups, helps, separators = parse_string(xml_string)
        except Exception, e:
            raise ConfigError, "erreur au chargement d'une configuration : %s" % e
        return self._read(new_files, new_variables, new_families, new_order, 
                          new_constraints, new_groups, helps, separators)

    def _read(self, new_files, new_variables, new_families, new_order, new_constraints, new_groups, helps, separators):
        """charge les données venant d'un dictionnaire"""
        # on commence par vérifier qu'il n'y a pas de doublons (variables)
        for varname in new_variables.keys():
            if self.variables.has_key(varname):
                # on de devrait pas avoir 2 fois la meme variable
                raise ConfigError, "deux variables ont le meme nom (%s)" % varname

        # ajout des nouveaux fichiers dans le dictionnaire final
        for fic, value in new_files.items():
            template = Template(fic, value['source'],
                                templatedir=self.templatedir,
                                mode=value['mode'], 
                                owner=value['owner'],
                                group=value['group'],
                                mkdir=value['mkdir'],
                                log=self.log.filename)
            self.files[fic] = {'var':template, 'hidden':value['hidden']}
            if value['filelist'] != '':
                self.filelists.setdefault(value['filelist'], []).append(fic)

        # ajout des nouvelles familles
        for family, data in new_families.items():
            if data['mode'] != 'expert' and family not in self.normal_families:
                # gestion de la liste des familles "normales"
                # -> normal un jour, normal toujours
                self.normal_families.append(family)
            # si famille inexistant, on l'ajoute
            if not self.families.has_key(family):
                # la famille est définie
                self.families[family] = data
            else:
                self.families[family]['vars'].extend(data['vars'])
                
        # stockage de l'odre des variables
        for label in new_order:
            if label in self.family_order:
                new_order.remove(label)
        self.family_order.extend(new_order)

        # ajout des nouvelles variables dans le dictionnaire final
        for key, values in new_variables.items():
            self.variables[key] = self._create_var(key, values, new_constraints)
        # si des contraintes sont présentes pour des variables définies dans un autre dictionnaire
        # elles sont affectées une fois les variables disponibles
        for type_const in ['checks', 'fills', 'auto']:
            # on vérifie les contraintes orphelines précédentes
            for var, liste_consts in self.orphan_constraints.get(type_const, {}).items():
                if var in new_variables:
                    self.update_constraints(var, type_const, liste_consts)
                    del(self.orphan_constraints[type_const][var])
            # on stocke les nouvelles contraintes orphelines
            for var, liste_consts in new_constraints.get(type_const, {}).items():
                if var not in new_variables:
                    if var not in self.variables:
                        # variables externe non chargée: on stocke
                        self.orphan_constraints[type_const].setdefault(var, []).extend(liste_consts)
                    else:
                        # contraintes à affecter à une variable présente avant chargement de ce fichier
                        self.update_constraints(var, type_const, liste_consts)

        # ajout des contraintes
        for var, conditions in new_constraints['conditions'].items():
            self.constraints.setdefault(var, []).extend(conditions)
        for var, auto_funcs in new_constraints['auto'].items():
            for func in auto_funcs:
                for param in func[1]:
                    if param['name'] == 'parametre':
                        self.auto_vars.setdefault(param['value'], []).append(var)

        # ajout des groupes
        for key, values in new_groups.items():
            if self.groups.has_key(key):
                # extension d'un groupe existant
                self.groups[key].extend(values)
                #raise ConfigError, "groupe déjà existant : %s" % key
            else:
                self.groups[key] = values

        # stockage de l'aide
        for key, values in helps['variables'].items():
            if self.helps['variables'].has_key(key):
                raise ConfigError, "aide déjà existante pour : %s" % key
            else:
                self.helps['variables'][key] = values
        for key, values in helps['families'].items():
            if self.helps['families'].has_key(key):
                raise ConfigError, "aide déjà existante pour : %s" % key
            else:
                self.helps['families'][key] = values

        # séparateurs (affichage)
        for var, value in separators.items():
            if self.separators.has_key(var):
                raise ConfigError, "plus d'un séparateur défini pour : %s" % var
            else:
                self.separators[var] = value

        # on revérifie toutes les contraintes de condition après chargement
        for var in self.constraints.keys():
            try:
                self.check_conditions(var)
            except KeyError:
                # on laisse passer si la cible n'est pas présente
                # (permet de mettre une condition sur un élément d'un autre
                # dictionnaire sans se soucier de l'ordre de chargement)
                traceback.print_exc()
                pass

 

    def read_dir(self, dir_config):
        """
        lecture d'un répertoire entier de dictionnaires
        """
        # chargement des dictionnaires
        if type(dir_config) != list:
            dir_config = [dir_config]
        for dir in dir_config:
            if isdir(dir):
                fichiers = listdir(dir)
                fichiers.sort()
                for dico in fichiers:
                    if dico.endswith('.xml'):
                        # lecture du fichier de configuration
                        try:
                            self.read(join(dir, dico))
                        except Exception, err:
                            sys.exit('erreur lors du chargement de %s : %s' % (join(dir, dico), str(err)))

        
    def _create_var(self, name, var, constraints):
        # recherche des fonctions liées à cette variable
        checks = []
        fills = []
        auto = []
        if constraints['checks'].has_key(name):
            checks.extend(constraints['checks'][name])
        if constraints['fills'].has_key(name):
            fills.extend(constraints['fills'][name])
        if constraints['auto'].has_key(name):
            auto.extend(constraints['auto'][name])
        eolevar = type_factory(
                       context = self,
                       mime = var['type'],
                       name = name,
                       valeole = var['value'],
                       description = var['description'],
                       hidden = var['hidden'],
                       multi = var['multi'],
                       checks = checks,
                       fills = fills,
                       auto = auto
                       )

        return eolevar
        
    def update_constraints(self, var, type_const, liste_consts):
        if hasattr(self.variables.get(var), type_const):
            getattr(self.variables[var], type_const).extend(liste_consts)

    def set_value(self, var, values, default=False, force=False):
        # actualisation de la valeur
        self.variables[var].set_value(values, default, force=force)
        # vérification des dépendances
        self.check_conditions(var, force)
 
    def get_value(self, var, default=False):
        # lecture de la valeur
        val = self.variables[var].get_value(default)
        if val == []:
            return ['']
        else:
            return val

    def get_final_value(self, var):
        return self.variables[var].get_final_value()


    def get_prec_value(self, var):
        val = self.variables[var].get_prec_value()
        if val == []:
            return ['']
        else:
            return val

    def get_description(self, var):
        return self.variables[var].description
 
    def instance(self, debug=False):
        """
        Instancie les fichiers selons les valeurs des variables eoles
        """
        self.log.info('===========================================')
        self.log.info('        Debut de l\'instanciation')
        self.log.info('===========================================')
        for fic, value in self.files.items():
            if value['hidden'] == False:
                try:
                    value['var'].verify()
                except Exception,e:
                    print e
                    sys.exit(1)
        for fic, value in self.files.items():
            if value['hidden'] == False:
                value['var'].process(self)
            else:
                self.log.info('Fichier %s non instancie'%fic)
                self.log.info('-------------------------------------------')
        self.log.info('===========================================')
        self.log.info('        Fin de l\'instanciation')
        self.log.info('===========================================')

    def load_values(self, store_file="/etc/eole/config.eol"):
        """initialise les valeurs des variables depuis un fichier cfg"""
        # vérification des groupes de variables ?
        try:
            self.store.read(store_file)
        except:
            return False
        self._load_values()

    def _load_values(self, force=False):
        # faire des vérifications au chargement ?
        for varname in self.variables.keys():
            var = self.variables[varname]
            if self.store.has_section(varname):
                var.val = self.store.get(varname, 'val')
                # hack pour compenser le fait que ConfigParser renvoie une chaine après lecture dans un fichier
                if type(var.val) != list:
                    var.val = eval(var.val)
                var.valprec = self.store.get(varname, 'valprec')
                if type(var.valprec) != list:
                    var.valprec = eval(var.valprec)
                var.valdefault = self.store.get(varname, 'valdefault')
                if type(var.valdefault) != list:
                    var.valdefault = eval(var.valdefault)
        if force == False:
            for varname in self.variables.keys():
                self.check_conditions(varname)
    
        return True
    def reset_defaults(self):
        """recharge les valeurs par défaut du dictionnaire"""
        for key in self.variables:
            self.variables[key].val = self.variables[key].get_value(default=True)
                    
    def check_conditions(self, varname, force=False):
        """vérifie les contraintes imposées à la variable
        @param force: si force==True, on ne tient pas compte des valeurs obligatoires
        (utile pour saisir les valeurs par défaut d'une variante ou d'un module)"""
        # calcul des variables cachées
        if self.constraints.has_key(varname):
            for val in self.variables[varname].get_value():
                for func in self.constraints[varname]:
                    active = True
                    try:
                        params = [{'value':val}]
                        params.extend(func[5])
                        if not (func[0] == 'obligatoire' and force == True):
                            self.variables[varname].eval_func(func[0], params)
                    except Exception, e:
                        # on désactive les variables liées
                        active = False
                    # si la cible est connue, on l'active/desactive
                    for family_target in func[1]:
                        if family_target in self.families:
                            self.families[family_target]['hidden'] = not active
                    for target in func[2]:
                        if target in self.variables:
                            self.variables[target].set_active(active)
                    for file_target in func[3]:
                        if file_target in self.files:
                            self.files[file_target]['hidden'] = not active
                    for filelist_target in func[4]:
                        if self.filelists.has_key(filelist_target):
                            for tmpl in self.filelists[filelist_target]:
                                self.files[tmpl]['hidden'] = not active
        # calcul des variables automatiques
        if self.auto_vars.has_key(varname):
            for autovar in self.auto_vars[varname]:
                var = self.variables[autovar]
                try:
                    val = var.get_calculated_value(auto=True)
                except DependencyError:
                    # la valeur auto n'est pas active
                    pass
                    # var.val = var.valprec
                else:
                    var.set_value(val)


    def save_values(self, store_file="/etc/eole/config.eol", force=False):
        """sauvegarde les valeurs des variables"""
        for varname in self.variables.keys():
            force_var = force
            var = self.variables[varname]
            for family_name, family in self.families.items():
                if varname in family['vars']:
                    if family['hidden'] == True:
                        # famille hidden : on ne vérifie pas 'obligatoire'
                        force_var = True
                        break
            # la variable est hidden (ou une famille la contenant)
            # on prend la valeur par défaut si pas de valeur
            if var.val == [] or var.val == ['']:
                var.val = var.get_value(default=True)
            # vérification de la valeur
            var.check_value(var.val, force_var)
            # passage des chaines unicode en str
            for list_val in [var.val, var.valprec, var.valdefault]:
                for ind in range(len(list_val)):
                    if type(list_val[ind]) == unicode:
                        list_val[ind] = list_val[ind].encode(config.charset)

            if not self.store.has_section(varname):
                self.store.add_section(varname)
            self.store.set(varname, 'val', var.val)
            self.store.set(varname, 'valprec', var.valprec)
            self.store.set(varname, 'valdefault', var.valdefault)
        # sauvegarde du fichier
        try:
            file_conf = file(store_file, 'w')
            self.store.write(file_conf)
            file_conf.close()
        except:
            return False

        return True
    
    def get_files(self):
        """
        @return: la liste des fichiers cibles
        """
        return self.files

    def get_variables(self):
        """
        @return: dictionnaire des variables et leur valeur
        """
        return self.variables

    def get_families(self):
        """
        @return: dictionnaire des familles et fonctions associées
        """
        return self.families

    def get_normal_families(self):
        """
        @return: dictionnaire des familles NON EXPERT et fonctions associées
        """
        families = {}
        # FIXME : optimisation ?
        for family in self.normal_families:
            # on ne garde que les clés valides
            families[family] = self.families[family]
        return families

    def get_groups(self):
        """
        @return: dictionnaire des groupes
        """
        return self.groups

    def get_variable(self, name):
        """
        @return: la valeur de la variable name
        """
        return self.variables[name]

    def get_constraints(self):
        """
        @return: liste des contraintes
        """
        return self.constraints
       

if __name__ == "__main__":
    pass
