# -*- coding: UTF-8 -*-
###########################################################################
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
#  Fonctions de calcul et validation des variables Creole
#
###########################################################################
from commands import getstatusoutput
import subprocess
import os, sys, re, glob
from creole.config import f_version, func_dir
from creole.error import TypeEoleError, DependencyError, FileNotFound, ConfigError
from creole.utils import classes
from creole import ifconfig
import imp


# type paramètre
class Param:
    """paramètre pour les fonctions de contrainte"""
    def __init__(self, value, name=None, typ=None):
        self.value = value
        self.name = name
        self.type = typ


#############################
## Fonctions de vérification
#############################

def valid_ip(data):
    """fonction de validation de la saisie d'une adresse ip"""
    
    # variable non obligatoire
    if data == "":
        return True
    
    # découpage de l'ip
    ip = data.split('.')
    if len(ip) != 4:
        raise TypeEoleError("mauvaise syntaxe d'adresse ip")
    
    # vérification des nombres
    for part in ip:
        try:
            num = int(part)
        except ValueError:
            raise TypeEoleError("l'ip doit être constituée de 4 nombres séparés par des points")
        else:
            if not (0 <= num < 256):
                raise TypeEoleError("les nombres doivent etre compris entre 0 et 255")
        
    # on n'a pas rencontré d'erreur        
    return True

def valid_intervalle(data):
    """ fonction de validation d'une plage d'ip """
    if data == "":
        return True
    
    liste = data.split()
    if len(liste) != 2:
        raise TypeEoleError("vous devez spécifier 2 adresses ip")
    else:
        # on vérifie la syntaxe des ips
        for ip in liste:
            valid_ip(ip)
    return True

def valid_network(data, ip=None):
    """fonction de validation de la saisie d'un réseau"""
    if data == "":
        return True
    
    # on utilise la fonction de test d'adresse ip pour la syntaxe
    valid_ip(data)
    if ip is not None:
        # on effectue les tests spécifiques au réseau (fonction de l'ip ?)
        ip_parts = ip.split(".")
        network_parts = data.split(".")
        for i in range(4):
            if (int(network_parts[i]) != 0) and (network_parts[i] != ip_parts[i]):
                raise TypeEoleError("le réseau spécifié ne correspond pas à l'ip "+ip)
    
    # aucune erreur rencontrée
    return True

def valid_netmask(data):
    """fonction de validation de la saisie d'un masque réseau"""
    if data == "":
        return True
    
    # on utilise la fonction de test d'adresse ip pour la syntaxe
    valid_ip(data)
    # on effectue les tests spécifiques aux masques réseaux
    netmask = []
    for part in data.split("."):
        netmask.append(int(part))
    error = 0
    mask_len = 0
    done = 0
    for part in netmask:
        bitmask = 0x80
        for bit in range(8):
            if not done:
                if part & bitmask:
                    mask_len = mask_len + 1
                else:
                    done = 1
            else:
                if part & bitmask:
                    error = 1
            bitmask = bitmask >> 1

    if error == 1:
        raise TypeEoleError("syntaxe de masque réseau invalide")
    # pas d'erreur rencontrée
    return True

def valid_booleen(data, type_bool="alpha"):
    """fonction de validation d'une réponse booléenne
       type_bool : alpha ou num (O/N ou 0/1)
    """
    if data == "":
        return True

    if (data[0].upper() != 'O') and (data[0].upper() != 'N'):
        raise TypeEoleError("vous devez repondre par 'O' ou 'N'")
    # pas d'erreurs détectées
    return True

def valid_enum(data, liste=['oui','non','Oui','Non','OUI','NON','o','n','O','N'], checkval="True"):
    """fonction de validation d'une réponse en fonction
    d'une liste de valeur possibles
    """
    if data == "":
        return True

    try:
        if type(liste) != list:
            liste = eval(liste)
    except:
        raise TypeEoleError("la liste des valeurs possibles n'est pas valide")
    if data not in liste and checkval != "False":
        raise TypeEoleError("repondez par un des choix suivants : "+str(liste))
    return True

def valid_regexp(data, exp_reg, err_msg="la syntaxe est incorrecte"):
    """fonction de validation d'une saisie par expression régulière"""
    if data == "":
        return True
    match = re.match(exp_reg, data)
    if match is None:
        raise TypeEoleError(err_msg)
    else:
        return True

def valid_entier(data, mini=None, maxi=None):
    """ fonction de validation d'un nombre (intervalle optionnel)"""
    if data == "":
        return True
    try:
        value = int(data)
    except ValueError:
        raise TypeEoleError("vous devez saisir un nombre")
    else:
        if mini is not None and value < int(mini):
            raise TypeEoleError("donnez un entier supérieur à "+str(mini))
        if maxi is not None and value > int(maxi):
            raise TypeEoleError("donnez un entier inférieur à "+str(maxi))
        return True

def valid_yesno(data):
    """Yes ou No pour les fichiers de config"""
    if data not in ["Yes", "No"]:
        raise TypeEoleError("la valeur doit être Yes ou No")
    return True

def valid_system(data, cmd, result, *args):
    """ test d'une commande système """
    if data == "":
        return True
    
    cmd_string = cmd
    for arg in args:
        cmd_string += " "+arg
    
    # exécution de la fonction
    code_retour = os.system(cmd_string)
    
    if code_retour == result:
        return True
    else:
        raise TypeEoleError("la commande "+cmd_string+" a echoué")

def valid_differ(data, value):
    """ test d'une valeur indésirable """
    if data == value:
        raise TypeEoleError("la valeur doit être différente de %s"%value)
    else:
        return True

def valid_chaine(data, forbidden_chars=[' ']):
    """Recherche de caractèresinterdits dans une chaine"""
    forbidden_chars = eval(forbidden_chars)
    for char in forbidden_chars:
        if char in data:
            raise TypeEoleError("les caractères (%s) sont interdits"%','.join(forbidden_chars))
    return True

def valid_phpmyadmin(data, apache):
    """ phpmyadmin dépend d'apache """
    if data == 'oui' and apache == 'non':
        raise TypeEoleError("L'activation de PhpMyAdmin nécessite celle d'Apache")
    return True

def obligatoire(data):
    """ fonction qui vérifie que les données ne sont pas vides
    """
    if data == [] or data == "" or data[0].strip() == "":
        raise TypeEoleError("ce paramètre est obligatoire")
    else:
        return True

################################################################
## Fonctions de calcul de dépendance entre variables et groupes 
################################################################

def hidden_if_not_in(val, *args):
    """ on cache si la valeur n'est pas présente dans celles proposées """
    if val not in args:
        raise DependencyError("""%s n'est pas dans %s""" % (val, args))

def hidden_if_in(val, *args):
    """ on cache si la valeur est présente dans celles proposées """
    if val in args:
        raise DependencyError("""%s est dans %s""" % (val, args))

# TODO : fonctions du genre range() ??

#############################################
## Fonctions de saisie et calcul des valeurs
#############################################

def calc_network(ip, netmask):
    """ calcule le réseau par défaut à partir d'un masque et d'une ip """
    # si les ip et netmask viennent du dictionnaire, on les récupère
    if ip == "" or netmask == "":
        return ""
    broadcast, network = calc_adresses(ip, netmask)
            
    return network

def calc_broadcast(ip, netmask):
    """ calcule l'adresse de broadcast par défaut à partir d'un masque et d'une ip """
    # si les ip et netmask viennent du dictionnaire, on les récupère
    if ip == "" or netmask == "":
        return ""
    broadcast, network = calc_adresses(ip, netmask)
            
    return broadcast

def create_ip(ip_list):
    """ création d'une adresse ip à partir d'une liste"""
    template = """%i.%i.%i.%i"""
    return template % tuple(ip_list)
    
def create_ip_list(ip):
    """ création d'une liste à partir d'une adresse ip"""
    ipa, ipb, ipc, ipd = ip.split('.')
    return [int(ipa), int(ipb), int(ipc), int(ipd)]

def calc_adresses(ip, mask):
    """calcul l'adresse de broadcast à partir de l'ip et du netmask
    """
    mask_list = create_ip_list(mask)
    ip_list = create_ip_list(ip)
    broadcast_list = []
    network_list = []
    for i in range(4):
        #pour l'adresse réseau c'est simple
        network_list.append(int(ip_list[i]) & int(mask_list[i]))
        #broadcast_part=255 si mask[i]=255
        for j in range(256):
            if mask_list[i]&j == 0:
                broadcast_part = j
        #calcul du masque spécial
        if broadcast_part != 255:
            broadcast_part += mask_list[i]&ip_list[i]
        broadcast_list.append(broadcast_part)
    broadcast = create_ip(broadcast_list)
    network = create_ip(network_list)
    return broadcast, network

def calc_classe(netmask):
    """calcule la classe du réseau à partir du netmask"""

    if classes.has_key(netmask):
        return(classes[netmask])
    else:
        return(netmask)

def calc_netmask(plage):
    """calcule le reseau a partir de sa classe"""
    if plage in classes.values():
        for mask, plaj in classes.items():
            if plaj == plage:
                return mask
    else:
        return plage

def calc_pg_cache_size():
    """
    Calcul la taille du cache postgreql en fonction de la RAM
    """
    procmem = open('/proc/meminfo')
    meminfo = procmem.read().split('\n')
    procmem.close()
    for line in meminfo:
        if line.startswith('MemTotal'):
            mem_tot = float(line.split()[1])
            break
    mem_cache = int( (mem_tot * 0.72) / 8 )
    return str(mem_cache)

def calc_val(valeur):
    """ valeur par défaut "en dur" """
    return(valeur)

def get_home_path():
    """Récupère le chemin d'accès au homedir
    (/home par défaut, /partages sur OpenVZ)
    """
    if os.system('dpkg --status eolebase-openvz &>/dev/null') == 0:
        return '/partages'
    else:
        return '/home'
    
#################################################
## Fonctions de calcul des valeurs automatiques
#################################################

def calc_network_auto(ip, mask):
    """calcul l'adresse de broadcast à partir de l'ip et du netmask
    """
    mask_list = create_ip_list(mask)
    ip_list = create_ip_list(ip)
    network_list = []
    for i in range(4):
        #pour l'adresse réseau c'est simple
        network_list.append(int(ip_list[i]) & int(mask_list[i]))
    network = create_ip(network_list)
    return network

def server_mem():
    """renvoit la mémoire du serveur
    """
    procmem = open('/proc/meminfo')
    meminfo = procmem.read().split('\n')
    procmem.close() 
    for line in meminfo:
        if line.startswith('MemTotal'):
            mem_tot = float(line.split()[1])
            break
    return str(mem_tot)

def kernel_version():
    """ renvoit la version du kernel
    """
    return_code, output = getstatusoutput('uname -r')
    if return_code == 0:
        return str(output)
    else:
        return ''

def auto_eth(nom_carte, **kw):
    """ renvoit l'adresse de la carte nom_carte
    """
    if kw['condition'].lower() == kw['parametre'].lower():
        eths = ifconfig.ifconfig()
        if nom_carte in eths.getInterfaceList():
            eth = eths.getAddr(nom_carte)
        else:
            return ''
        return str(eth)
    else: 
        raise DependencyError

def auto_netmask(nom_carte, **kw):
    """ renvoit le netmask pour la carte nom_carte
    """
    if kw['condition'].lower() == kw['parametre'].lower():
        eths = ifconfig.ifconfig()
        if nom_carte in eths.getInterfaceList():
            netmask = eths.getNetmask(nom_carte)
        else:
            return ''
        return str(netmask)
    else:
        raise DependencyError

def auto_network(nom_carte, **kw):
    """ renvoit le network pour la carte nom_carte
    """
    if kw['condition'].lower() == kw['parametre'].lower():
        eths = ifconfig.ifconfig()
        if nom_carte in eths.getInterfaceList():
            ip = eths.getAddr(nom_carte)
            netmask = eths.getNetmask(nom_carte)
            network = calc_network_auto(ip, netmask)
        else:
            return ''
        return str(network)
    else:
        raise DependencyError

def auto_broadcast(nom_carte, **kw):
    """ renvoit le broadcast pour la carte nom_carte
    """
    if kw['condition'].lower() == kw['parametre'].lower():
        eths = ifconfig.ifconfig()
        if nom_carte in eths.getInterfaceList():
            broadcast = eths.getBroadcast(nom_carte)
        else:
            return ''
        return str(broadcast)
    else:
        raise DependencyError

def auto_module():
    """ renvoit le nom du module Eole
    """
    # FIXME : doublon avec fonctionseole 
    # mais pb d'import croisé !
    if not os.path.isfile(f_version):
        raise FileNotFound, '%s non trouvé' % f_version
    f_vers = open(f_version)
    content = f_vers.read().strip()
    f_vers.close()
    if content == '':
        raise ConfigError, '%s vide' % f_version
    return content.split('-')[0]

####################################################
# fonctions diverses accessibles dans les templates
####################################################

def list_files(rep, wildcards='*', stripext=False, basename=True, default=[]):
    """ liste des fichiers dans un répertoire """
    res = default
    if os.path.isdir(rep):
        res = []
        for fic in glob.glob(rep+'/%s' % wildcards):
            if basename:
                fic = os.path.basename(fic)
            if stripext:
                fic = os.path.splitext(fic)[0]
            res.append(fic)
        res.sort()
    return res

def is_file(filename):
    """ filtre permettant de verifier l'existence 
    d'un fichier avant de l'inclure par exemple """
    return os.path.isfile(filename)

def list(lst, sep=' '):
    """retourne une liste concatenee de donnees de la liste"""
    return sep.join(lst)

def purge_list(lst):
    """supprime les doublons dans une liste"""
    res = []
    for elt in lst:
        if elt not in res:
            res.append(elt)
    return res

def to_iso(data):
    """ encode une chaine en iso"""
    try:
        return unicode(data, "UTF-8").encode("ISO-8859-1")
    except:
        return data

def is_installed(package):
    """ vérifie si un paquet est installé ou pas"""
    process = subprocess.Popen(args=["dpkg", "-l", package.strip()], 
              stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
    code = process.wait()
    if code != 0:
        # paquet non installé
        return False
    dpkg_info = process.stdout.read()
    pkg_status = dpkg_info.strip().split('\n')[-1].split()
    if pkg_status[0].lower() == 'pn':
        # ancien paquet purgé (status = pn)
        return False
    return True

def installed_version(package):
    """renvoit la version d'un paquet"""
    process = subprocess.Popen(args=["dpkg", "-l", package.strip()], 
              stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
    code = process.wait()
    if code != 0:
        # paquet non installé
        return ''
    dpkg_info = process.stdout.read()
    pkg_status = dpkg_info.strip().split('\n')[-1].split()
    if pkg_status[0].lower() == 'pn':
        # ancien paquet purgé (status = pn)
        return ''
    return pkg_status[2]

def custom_join(multivar, separator=' '):
    """ mise à plat d'une variable multiple"""
    return separator.join([i.value for i in multivar])

def certs_nuauth():
    """ liste des certificats pour nuauth """
    namecert = ''
    nuauth_tls_key = ''
    rep = '/etc/ipsec.d'
    repdefault='/etc/nufw/certs/'
    wildcards = '*.pem'
    if os.path.isdir(rep):
        if len(glob.glob(rep+'/%s' % wildcards)) > 0:
            namecert = os.path.basename(glob.glob(rep+'/%s' % wildcards)[0])
            nuauth_tls_key = os.path.join(rep, 'private', 'priv%s' % namecert)
            if not os.path.isfile(nuauth_tls_key):
                nuauth_tls_key = ''
            else:
                nuauth_tls_cert = os.path.join(rep, namecert)
                if not os.path.isfile(nuauth_tls_cert):
                    nuauth_tls_key = ''
                else:
                    nuauth_tls_cacert = os.path.join(rep, 'cacerts', 'CertifCa.pem')
                    if not os.path.isfile(nuauth_tls_cacert):
                        nuauth_tls_key = ''
    if nuauth_tls_key == '':
        nuauth_tls_key=os.path.join(repdefault,'nuauth-key.pem')
        if not os.path.isfile(nuauth_tls_key):
            nuauth_tls_key='none'
        nuauth_tls_cert=os.path.join(repdefault,'nuauth-cert.pem')
        if not os.path.isfile(nuauth_tls_cert):
            nuauth_tls_key='none'
            nuauth_tls_cert='none'
        nuauth_tls_cacert='none'
    res = [nuauth_tls_key, nuauth_tls_cert, nuauth_tls_cacert]
    return res

def key_nuauth():
    """retourne la clef privee pour nuauth"""
    return certs_nuauth()[0]

def cert_nuauth():
    """retourne le certificat pour nuauth"""
    return certs_nuauth()[1]

def cacert_nuauth():
    """retourne l'authorité de certification pour nuauth"""
    return certs_nuauth()[2]

#########################################################################
# Chargement de fonctions supplémentaires depuis /usr/share/creole/funcs
#########################################################################

dict_locals = locals()
for mod_file in glob.glob('%s/*.py' % func_dir):
    try:
        if os.path.basename(mod_file) != '__init__.py':
            mod_perso = imp.load_source('mod_perso', mod_file)
            for attr in dir(mod_perso):
                if not attr.startswith('__'):
                    if attr not in dict_locals:
                        dict_locals[attr] = getattr(mod_perso, attr)
                    else:
                        # on n'écrase pas les fonctions déjà présentes !
                        print "%s - déjà défini : %s" % (mod_file, attr)

    except Exception, e:
        print "erreur lors du chargement des fonctions personnalisées dans %s" % mod_file
        print str(e)

def main():
    """fonction de test  de la bibliothèque"""
    try:
        valid_netmask(sys.argv[1])
        print "ok"
    except Exception, msg:
        print "erreur : %s" % str(msg)

if __name__ == "__main__":
    main()
