# -*- coding: UTF-8 -*-
"""
Gestion du mini-langage de template 
On travaille sur les fichiers cibles
"""

from os.path import isfile, isdir, basename, join, dirname
from os import system, makedirs
import traceback
from logging import error

from Cheetah.Template import Template as ChtTemplate
from Cheetah.NameMapper import NotFound as CheetahNotFound

from creole.error import FileNotFound, TemplateError
from creole.log import logger
from creole import eosfunc
import creole.config as cfg

class CheetahTemplate(ChtTemplate):
    """classe pour personnaliser et faciliter la construction
    du template Cheetah
    """
    def __init__(self, filename, context):
        self.context = context
        eos = {}
        for func in dir(eosfunc):
            if not func.startswith('_'):
                eos[func] = getattr(eosfunc, func)
        ChtTemplate.__init__(self, file=filename, searchList=[context, eos, {'is_defined':self.is_defined}],
                             compilerSettings={'directiveStartToken' : '%',
                                               'cheetahVarStartToken' : '%%',
                                               'EOLSlurpToken' : '%',
                                               'PSPStartToken' : 'µµµµµµµµµµ',
                                               'PSPEndToken' : 'µµµµµµµµµµ',
                                               'commentStartToken' : 'µµµµµµµµµµ',
                                               'commentEndToken' : 'µµµµµµµµµµ',
                                               'multiLineCommentStartToken' : 'µµµµµµµµµµ',
                                               'multiLineCommentEndToken' : 'µµµµµµµµµµ'})

    def is_defined(self, varname):
        """filtre permettant de ne pas lever d'exception au cas où
        la variable ne serait pas définie
        """
        return varname in self.context

class Template:
    """
    Classe qui transforme un fichier template en fichier instanciable
    """
    def __init__(self, target, source='', mode='', owner='', group='', 
                 mkdir=False, templatedir=cfg.templatedir, log=None):
        """
        @param target   : nom long du fichier cible
        @param source : nom du fichier template si different du nom de la cible
        @param templatedir : répertoire des template
        @param mode : droits Unix à appliquer (ex : 0644)
        @param owner : propriétaire du fichier
        @param group : groupe Unix propriétaire du fichier
        @param mkdir : création du répertoire si nécessaire
        """
        # fichier cible
        self.target = target
        # fichier template
        if source:
            self.filename = join(templatedir, source)
        else:
            self.filename = join(templatedir, basename(self.target))
        # attributs Unix 
        self.mode = mode
        self.owner = owner
        self.group = group
        self.mkdir = mkdir
        self.log = logger
        self.var_dict = None

        
    def verify(self):
        """
        verifie que les fichiers existent
        """
        if not isfile(self.filename):
            raise FileNotFound, "le fichier %s n'existe pas" % self.filename
        if not isdir(dirname(self.target)):
            if not self.mkdir:
                raise FileNotFound, "le repertoire %r n'existe pas" % dirname(self.target)            
            makedirs(dirname(self.target))
        if not isfile(self.target):
            system('cp %s %s' % (self.filename, self.target))


    def __cmp__(self, other):
        """
        comparer deux objets templates
        en pratique : utiliser pour faire un sort sur les noms de fichiers
        """
        if isinstance(other, Template):
            return cmp(self.target, other.target)
        return 1


    def process(self, eoledict):
        """
        traitement du template 
        et écriture du fichier cible
        """
        # 1) remplacement des variables Eole
        if self.var_dict is None:
            self.var_dict = self._eoledict_to_dict(eoledict)
        try:
            cheetah_template = CheetahTemplate(str(self.filename), self.var_dict)
            data = str(cheetah_template)
        except CheetahNotFound, err:
            # XXX Fragile: err.args devrait avoir la
            #              forme ("cannot find 'varname'",)
            varname = err.args[0][13:-1]
            msg = "Erreur: Utilisation d'une variable non existante dans %s : %r" % (self.filename, varname)
            self.log.error(msg)
            raise TemplateError, msg
        except UnicodeDecodeError, err:
            msg = "Probleme d'encodage detecte dans le fichier %s" % self.filename
            self.log.error(msg)
            raise TemplateError, msg
        except Exception, err:
            msg = "Erreur lors de l'instanciation du template %s : %s" % (self.filename, str(err))
            self.log.error(msg)
            raise TemplateError, msg
        # 2) écriture du fichier cible
        try:
            file_h = file(self.target, 'w')
            file_h.write(data)
            file_h.close()
        except:
            msg = "Impossible d'ecrire dans le fichier %s" % self.target
            self.log.error(msg)
            raise FileNotFound, msg
        # 3) Gestion des attributs Unix
        if self.owner != '':
            system("chown %s %s" % (self.owner, self.target))
        if self.group != '':
            system("chgrp %s %s" % (self.group, self.target))
        if self.mode != '':
            system("chmod %s %s" % (self.mode, self.target))
        # resultat
        self.log.info('fichier %s :' % basename(self.target))
        self.log.info('fichier copie dans %s' % self.target)


    def _eoledict_to_dict(self, eoledict):
        """
        génération du tableau de remplacement
        @return : dictionnaire du type { 'variable' : VALEUR }
                  où VALEUR est soit une valeur directement utilisable
                  (entier, chaine, liste, etc.) soit une instance de
                  `MultiVariable` dans le cas d'une variable multiple.
        """
        return dict( (varname, eoledict.get_final_value(varname))
                     for varname in eoledict.variables )

