# -*- coding: UTF-8 -*-

"""
Objets Eole
"""

from creole import eosfunc
from creole.error import TypeEoleError, DependencyError
from creole.config import charset
from creole.utils import encode_list

def type_factory(mime, name, **args):
    """
    @param mime: le type mime du type recherché
    @param name: le nom de la variable eole à créer
    @return: un objet type correspondant au mimetype
    """
    if not isinstance(mime, basestring):
        raise TypeEoleError("mauvais type mime '%s' pour la variable %s " % (mime, name))
    if not mime.strip():
        raise TypeEoleError("type de variable est vide pour la variable " + name)
    
    mime = mime.lower()
    if mime not in REGISTERED_TYPES:
        raise TypeEoleError("type de variable inconnu ["+mime+"] pour la variable"+name)
    return REGISTERED_TYPES[mime](name, **args)

REGISTERED_TYPES = {}

# objet de base Eole
class autoregistrable(type):
    def __new__(mcs, name, bases, classdict):
        mime = classdict.get('mime')
        cls = type.__new__(mcs, name, bases, classdict)
        if mime:
            REGISTERED_TYPES[mime] = cls
        return cls


class MultiVariable:
    """
    >>> var = MultiVariable(12, dict(netmask='truc', network='muche'))
    >>> var
    12
    >>> var.netmask
    'truc'
    >>> var.network
    'muche'
    """
    def __init__(self, value, slavedict):
        self.value = value
        self.__dict__.update(slavedict)

    def __repr__(self):
        return str(self.value)

    def __eq__(self, other):
        # pour gérer les comparaisons Multivalué/chaine
        if hasattr(other,'value'):
            return self.value == other.value
        else:
            return self.value == other

class EoleVar:
    __metaclass__ = autoregistrable

    def __init__(self, name, context={}, description=None, val=[], valprec=[], valdefault=[], valeole=[], hidden=False, multi=False, checks=[], fills=[], auto=[]):
        """
        contient normalement les tests sur l'objet
        @param name: nom de la variable
        @param description: libellé de la variable
        @param val: valeur utilisée        
        @param valprec: valeur précédente
        @param valdefault: valeur par défaut régionale
        @param valeole: valeur par défaut d'origine
        @param checks: liste des fonction(s) de vérification supplémentaire
        @param fills: liste des fonction(s) de calcul de valeur par défaut
        @param auto: variable automatique ?
        """
        self.eole_dico = context
        if context != {}:
            self.context = context.variables
        else:
            self.context = {}
        self.auto = auto
        self.val = val
        self.valprec = valprec
        self.valdefault = encode_list(valdefault)
        self.valeole = encode_list(valeole)
        # autres paramètres
        self.name = name
        self.description = description
        self.hidden = hidden
        self.multi = multi
        self.checks = []
        self.fills = []
        # initialisation de la liste des fonctions
        if type(checks) == list and type(fills) == list:
            self.checks.extend(checks)
            self.fills.extend(fills)
        else:
            raise TypeEoleError("Erreur à l'initiatisation des fonctions")
    
    def __eq__(self, other):
        return self.name == other.name and \
               self.val == other.val and \
               self.valprec == other.valprec and \
               self.valdefault == other.valdefault and \
               self.valeole == other.valeole and \
               self.description == other.description
               
    def set_active(self, active):
        self.hidden = not active
       
       
    def set_value(self, vals, default=False, force = False):
        """affectation d'une valeur à la variable
        @param vals : valeurs
        @param default: assignation de la valeur par défaut si True
        """
        # si la valeur n'est pas dans un liste, on en crée une
        if type(vals) != list:
            vals=[vals]
        # vérification des valeurs (lève une exception si erreur)
        self.check_value(vals, force)
        if not default :
            if self.val != vals:
                # inutile d'écraser si idem
                self.valprec = []
                self.valprec.extend(self.val)
                self.val = vals
        else:
            self.valdefault = vals

    def check_value(self, vals, force = False):
        # parcours des couples (fonction, arguments)
        for couple in self.checks:
            try:
                # cas particulier : valeurs obligatoires
                if couple[0] == 'obligatoire':
                    if force == True or self.hidden == True:
                        # en mode forcé, on ignore les variables obligatoires
                        continue
                    eosfunc.obligatoire(vals)
                # les variables sont de type liste 
                # on doit appliquer la fonction sur chacune
                for val in vals:
                    valeurs=[{'value':val}]
                    if couple[1] != []:
                        valeurs.extend(couple[1])
                    # évaluation fonction avec les valeurs en 1er argument
                    self.eval_func(couple[0], valeurs)
            except TypeEoleError, e:
                # on précise le nom de la variable dans le message
                raise TypeEoleError, str(self.name) + " : " + e.args[0]

    def eval_func(self, function, params, index=0):
        # parcours des arguments
        arguments = ''
        for param in params:
            # gestion des keyword args
            if param.has_key('name'):
                if param['name'] != '':
                    arguments += '%s=' % param['name']
            value = param['value']
            # gestion des références aux variables eole
            if param.has_key('type'):
                if param['type'] == 'eole':
                    vals = self.context[param['value']].get_value()
                    if len(vals) == 0:
                        value = ''
                    else:
                        if self.is_linked(param['value']):
                            # la variable est groupée
                            value = vals[index]
                        else:
                            # sinon on prend toujours la première
                            value = vals[0]
                elif param['type'] == 'python':
                    value = eval(value,globals(),locals())
            # valeur du paramètre
            arguments += '"%s", ' % str(value)
            
        # les fonctions sont forcément dans eosfunc
        # on fait un eval car on n'a que le nom de la fonction 
        if type(function) is unicode:
            function = function.encode(charset)
        if type(arguments) is unicode:
            arguments = arguments.encode(charset)
        return eval( 'eosfunc.%s(%s)' % (function, arguments) )

    def is_linked(self, varname):
        for master, slaves in self.eole_dico.groups.items():
            if self.name == master or self.name in slaves:
                if varname == master or varname in slaves:
                    return True
        return False

    def get_value(self, default=False):
        """retourne la valeur actuelle
        @param default: force le retour de la valeur par défaut
        @return: valeur actuelle ou valeur par défaut
        (si pas de valeur par défaut, retour de la valeur eole)
        """
        # si la variable est automatique, on tente de renvoyer la valeur calculée si il y a
        if self.auto != []:
            try:
                return self.get_calculated_value(auto=True)
            except DependencyError:
                # on ne doit pas utiliser la valeur auto
                pass
        if not default and self.val != []:
            return self.val
        if self.valdefault != []:
            return self.valdefault
        else:
            if len(self.fills) > 0:
                return self.get_calculated_value()
            elif self.valeole != []:
                return self.valeole
            else:
                return []

    def get_final_value(self):
        """
        :return: valeur directement utilisable pour Cheetah
        """
        value = self.get_value()
        if value in ([],['']):
            return ''
        
        if self.multi:
            return list(self)
        else:
            # On utilise getitem pour gérer la transformation en
            # MultiVariable quand nécessaire
            return self.get_final_value_at_index(0)

    def get_final_value_at_index(self, index):
        """On renvoie la une valeur directement utilisable
        par Cheetah (avec transformation en MultiVariable si nécessaire)
        """
        value = self.get_value()[index]
        slaves = self.get_slaves()
        if slaves:
            slavedict = {}
            for varname, values in slaves.items():
                slavedict[varname] = values[index]
            return MultiVariable(value, slavedict)
        else:
            return value

    def get_prec_value(self):
        """retourne la valeur précédente
        """
        return self.valprec


    def get_calculated_value(self, index=None, auto=False):
        vals = []
        if auto == True:
            funcs = self.auto
        else:
            funcs = self.fills

        for fill in funcs:
            # calcul de la valeur
            func = fill[0]
            params = fill[1]
            # on regarde si cette variable est liée à self pour récupérer
            # le bon index de valeur
            grouped = 0
            for param in params:
                if param['type'] == 'eole':
                    if self.is_linked(param['value']):
                        grouped = len(self.context[param['value']].val)
                        break
            if grouped != 0:
                vals = []
                for i in range(grouped):
                    vals.append(self.eval_func(func, params, i))
            else:
                vals = [self.eval_func(func, params)]
        return vals


    def get_slaves(self):
        if self.multi:
            self_values = self.get_value()
        slaves = {}
        # on teste si self est une variable maître
        if self.name in self.eole_dico.groups:
            # on récupère le nom des variables esclaves
            for slave in self.eole_dico.groups[self.name]:
                # on récupère les valeurs pour cette variable
                slave_values = self.eole_dico.variables[slave].get_value()
                slaves[slave] = slave_values
                # si c'est multi-valué, il **faut** que les variables esclaves
                # aient le même nombre de valeurs que la variable maître
                # cas particulier : liste vide pour la variable maitre, mais des valeurs
                #                   par défaut existent dans les variables  esclaves.
                if self.multi and len(self_values) != 0:
                    assert len(slave_values) == len(self_values), \
                           "les variables esclaves doivent avoir le meme nombre d'elements" \
                           " que la variable maitre (%s)" % self.name
        return slaves

    def __getitem__(self, index):
        """On rend la variable indexable"""
        assert self.multi, "impossible d'utiliser la notation indexee si la " \
               "variable n'est pas multi-valuee"
        return self.get_final_value_at_index(index)

    def __iter__(self):
        """On rend la variable itérable pour pouvoir faire:
        for ip in iplist:
            print ip.network
            print ip.netmask
            print ip
        """
        assert self.multi, "impossible d'iterer sur une variable non multiple"
        if self.hidden == False:
            self_values = self.get_value()
            slaves = self.get_slaves()
            for index, value in enumerate(self_values):
                slavedict = {}
                for varname, values in slaves.items():
                    slavedict[varname] = values[index]
                # on rajoute un attribut "index" pour que l'indice de la valeur
                # soit accessible directement sur la variable, e.g. :
                # %for $ip in $liste_ips
                #  $ip.index # <=== ici !
                # %end for
                slavedict['index'] = index
                yield MultiVariable(value, slavedict)

# définition des types de variables Creole           

class TypedVar(EoleVar):
    checkers = []

    def __init__(self, name, **kwargs):
        EoleVar.__init__(self, name, **kwargs)
        self.checks[0:0] = self.checkers

class String(TypedVar):
    mime = 'string'

class Ip(TypedVar):
    mime = 'ip'
    checkers = [ ['valid_ip', []], ]
        
class Netmask(TypedVar):
    mime = 'netmask'
    checkers = [ ['valid_netmask', []], ]

class Number(TypedVar):
    mime = 'number'
    checkers = [ ['valid_entier', []], ]
 
class Boolean(TypedVar):
    mime = 'boolean'
    checkers = [ ['valid_booleen', []], ]
